/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "../WinQuake/quakedef.h"

//#include "../WinQuake/glquake.h"
//#include "../WinQuake/vid.h"
//#include "../WinQuake/sbar.h"
//#include "../WinQuake/sound.h"

#include "GLES/gl.h"
#include "GLES/egl.h"
#include "nanogl.h"
#include "wizGLES.h"

NativeWindowType hNativeWnd = 0;

viddef_t vid;				// global video state

static int scr_width, scr_height;
static float vid_gamma = 1.0;

qboolean is8bit = false;
qboolean fullsbardraw = true;
qboolean isPermedia;
qboolean gl_mtexable;

int min_vid_width;

EGLDisplay      glDisplay;
EGLConfig       glConfig;
EGLContext      glContext;
EGLSurface      glSurface;

unsigned short		d_8to16table[256];
unsigned		d_8to24table[256];
unsigned char		d_15to8table[65536];

void (*VID_SetGammaRamp)(unsigned short ramp[3][256]);

static cvar_t m_filter = {"m_filter", "0"};
int		texture_mode = GL_NEAREST;


int		texture_extension_number = 1;

float		gldepthmin, gldepthmax;

cvar_t	gl_ztrick = {"gl_ztrick","0"};

const char *gl_vendor;
const char *gl_renderer;
const char *gl_version;
const char *gl_extensions;

void	VID_SetPalette (unsigned char *palette)
{
	byte	*pal;
	unsigned r,g,b;
	unsigned v;
	int     r1,g1,b1;
	int k;
	//int		j,k,l,m;
	unsigned short i;
	unsigned	*table;
	//FILE *f;
	//char s[255];
	int dist, bestdist;

//
// 8 8 8 encoding
//
	pal = palette;
	table = d_8to24table;
	for (i=0 ; i<256 ; i++)
	{
		r = pal[0];
		g = pal[1];
		b = pal[2];
		pal += 3;
		
		v = (255<<24) + (r<<0) + (g<<8) + (b<<16);
		*table++ = v;
	}
	d_8to24table[255] &= 0xffffff;	// 255 is transparent

	for (i=0; i < (1<<15); i++) {
		/* Maps
		000000000000000
		000000000011111 = Red  = 0x1F
		000001111100000 = Blue = 0x03E0
		111110000000000 = Grn  = 0x7C00
		*/
		r = ((i & 0x1F) << 3)+4;
		g = ((i & 0x03E0) >> 2)+4;
		b = ((i & 0x7C00) >> 7)+4;
		pal = (unsigned char *)d_8to24table;
		for (v=0,k=0,bestdist=10000*10000; v<256; v++,pal+=4) {
			r1 = (int)r - (int)pal[0];
			g1 = (int)g - (int)pal[1];
			b1 = (int)b - (int)pal[2];
			dist = (r1*r1)+(g1*g1)+(b1*b1);
			if (dist < bestdist) {
				k=v;
				bestdist = dist;
			}
		}
		d_15to8table[i]=(unsigned char)k;
	}
}
void	VID_ShiftPalette (unsigned char *palette)
    {
 //    VID_SetPalette(palette);
    }
    
void CheckMultiTextureExtensions(void) 
{	
  //qglMTexCoord2fSGIS = (lpMTexFUNC)eglGetProcAddress("glMultiTexCoord2fARB");
 //qglSelectTextureSGIS = (lpSelTexFUNC)eglGetProcAddress("glActiveTextureARB");
  //if (qglMTexCoord2fSGIS && qglSelectTextureSGIS)
  //{   
  //  Con_Printf("Multitexture extensions found.\n");
  //  gl_mtexable = true;
  //}
  //else
  //{
    gl_mtexable = false;
  //}
}

EGLint attrib_list_fsaa[] =
{
  EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
  EGL_BUFFER_SIZE,  0,
  EGL_DEPTH_SIZE,   16,
  EGL_SAMPLE_BUFFERS, 1,
  EGL_SAMPLES,        4,
  EGL_NONE
};

EGLint attrib_list[] =
{
  EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
  EGL_BUFFER_SIZE,  0,
  EGL_DEPTH_SIZE,   16,
  EGL_NONE
};


void GL_Init (void)
{	
	EGLint numConfigs;
	EGLint majorVersion;
	EGLint minorVersion;
	
	glDisplay = eglGetDisplay( (NativeDisplayType)0 );
	if( glDisplay == EGL_NO_DISPLAY )
	{
		printf( "GL No Display failed\n" );
	}
	
	if( !eglInitialize( glDisplay, &majorVersion, &minorVersion ) ) 
	{
		printf( "GL Init failed\n" );
	}
	if( !eglChooseConfig( glDisplay, attrib_list, &glConfig, 1, &numConfigs ) )
	{
		printf( "GL Config failed\n" );
	}
	glContext = eglCreateContext( glDisplay, glConfig, NULL, NULL );
	if( glContext==0 )
	{
		printf( "GL Context failed\n" );
	}
	glSurface = eglCreateWindowSurface( glDisplay, glConfig, hNativeWnd, NULL );
	if( glSurface==0 )
	{
		printf( "GL Surface failed\n" );
	}
	printf( "EGL Init Completed\n" );
	
	eglMakeCurrent( glDisplay, glSurface, glSurface, glContext );    

	gl_vendor = (const char*)glGetString (GL_VENDOR);
	printf("GL_VENDOR: %s\n", gl_vendor);
	gl_renderer = (const char*)glGetString (GL_RENDERER);
	printf("GL_RENDERER: %s\n", gl_renderer);

	gl_version = (const char*)glGetString (GL_VERSION);
	printf("GL_VERSION: %s\n", gl_version);
	gl_extensions = (const char*)glGetString (GL_EXTENSIONS);
	printf("GL_EXTENSIONS: %s\n", gl_extensions);

	fullsbardraw = true;
	CheckMultiTextureExtensions ();
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
	glClearColor (0,0,0,0);
	glCullFace(GL_FRONT);
	glEnable(GL_TEXTURE_2D);

	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.666);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel (GL_FLAT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

/*
=================
GL_BeginRendering

=================
*/
void GL_BeginRendering (int *x, int *y, int *width, int *height)
    {
	//extern cvar_t gl_clear;

	*x = *y = 0;
	*width = vid.width;
	*height = vid.height;
    }


void GL_EndRendering (void)
{
    //S_ExtraUpdate ();
    eglSwapBuffers(glDisplay,glSurface);
    //S_ExtraUpdate ();

    if (fullsbardraw)
	    Sbar_Changed();
}

extern qboolean VID_Is8bit(void)
{
    return is8bit;
}

void VID_Init8bitPalette(void) 
{
/*	// Check for 8bit Extensions and initialize them.
	int i;
	void *prjobj;

	if ((prjobj = dlopen(NULL, RTLD_LAZY)) == NULL) {
		Con_Printf("Unable to open symbol list for main program.\n");
		return;
	}

	if (strstr(gl_extensions, "3DFX_set_global_palette") &&
		(qgl3DfxSetPaletteEXT = dlsym(prjobj, "gl3DfxSetPaletteEXT")) != NULL) {
		GLubyte table[256][4];
		char *oldpal;

		Con_SafePrintf("8-bit GL extensions enabled.\n");
		glEnable( GL_SHARED_TEXTURE_PALETTE_EXT );
		oldpal = (char *) d_8to24table; //d_8to24table3dfx;
		for (i=0;i<256;i++) {
			table[i][2] = *oldpal++;
			table[i][1] = *oldpal++;
			table[i][0] = *oldpal++;
			table[i][3] = 255;
			oldpal++;
		}
		qgl3DfxSetPaletteEXT((GLuint *)table);
		is8bit = true;

	} else if (strstr(gl_extensions, "GL_EXT_shared_texture_palette") &&
		(qglColorTableEXT = dlsym(prjobj, "glColorTableEXT")) != NULL) {
		char thePalette[256*3];
		char *oldPalette, *newPalette;

		Con_SafePrintf("8-bit GL extensions enabled.\n");
		glEnable( GL_SHARED_TEXTURE_PALETTE_EXT );
		oldPalette = (char *) d_8to24table; //d_8to24table3dfx;
		newPalette = thePalette;
		for (i=0;i<256;i++) {
			*newPalette++ = *oldPalette++;
			*newPalette++ = *oldPalette++;
			*newPalette++ = *oldPalette++;
			oldPalette++;
		}
		qglColorTableEXT(GL_SHARED_TEXTURE_PALETTE_EXT, GL_RGB, 256, GL_RGB, GL_UNSIGNED_BYTE, (void *) thePalette);
		is8bit = true;
	}
	
	dlclose(prjobj);*/
}

static void Check_Gamma (unsigned char *pal)
{
	float	f, inf;
	unsigned char	palette[768];
	int		i;

	if (i = COM_CheckParm("-gamma")) 
        {
		vid_gamma = Q_atof(com_argv[i+1]);
        }
	else
        {
		vid_gamma = 0.75f;
        }

	printf( "Gamma is %f\n", vid_gamma );

	for (i=0 ; i<768 ; i++)
	{
		f = pow ( (pal[i]+1)/256.0 , vid_gamma );
		inf = f*255 + 0.5;
		if (inf < 0)
			inf = 0;
		if (inf > 255)
			inf = 255;
		palette[i] = inf;
	}

	memcpy (pal, palette, sizeof(palette));
}

extern void VID_Init(unsigned char *palette)
{
    //int i;
    char	gldir[MAX_OSPATH];

    long width = 320;
    long height = 240;

    nanoGL_Init();

    // Create native window.
    printf( "VID_Init: Creating the window\n" ); 
    hNativeWnd = OS_CreateWindow();	                  
    if(!hNativeWnd)
    printf( "VID_Init: OS_CreateWindow Failed\n" ); 


//	Cvar_RegisterVariable (&vid_mode);
//	Cvar_RegisterVariable (&in_mouse);
//	Cvar_RegisterVariable (&in_dgamouse);
    Cvar_RegisterVariable (&m_filter);
    Cvar_RegisterVariable (&gl_ztrick);

    vid.maxwarpwidth = vid.width = vid.conwidth = width;
    vid.maxwarpheight = vid.height = vid.conheight = height;
    vid.aspect = 1.0;
    vid.numpages = 1;
    vid.colormap = host_colormap;
    vid.fullbright = 256 - LittleLong (*((int *)vid.colormap + 2048));
    vid.rowbytes = vid.conrowbytes = width;
    vid.recalc_refdef = true;
    min_vid_width = width;

    printf( "VID_Init: GL_Init starting\n" ); 
    GL_Init();
    printf( "VID_Init: GL_Init finished\n" );

    sprintf (gldir, "%s/glquake", com_gamedir);
    //Sys_mkdir (gldir);
    Check_Gamma(palette);
    VID_SetPalette (palette);

    // Check for 3DFX Extensions and initialize them.
    VID_Init8bitPalette();
}



void VID_Shutdown (void)
{
	eglMakeCurrent( glDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT );
	eglDestroySurface( glDisplay, glSurface ); 
	eglDestroyContext( glDisplay, glContext );
	eglTerminate( glDisplay );  

	nanoGL_Destroy();    
}



extern void VID_Changed()
    {
    }


/*
================
D_BeginDirectRect
================
*/
void D_BeginDirectRect() // (int /*x*/, int /*y*/, byte */*pbitmap*/, int /*width*/, int /*height*/)
{
}


/*
================
D_EndDirectRect
================
*/
void D_EndDirectRect() // (int /*x*/, int /*y*/, int /*width*/, int /*height*/)
{
}


/*
================
VID_LockBuffer
================
*/
//void VID_LockBuffer( void )
//{
//}


/*
================
VID_UnlockBuffer
================
*/
//void VID_UnlockBuffer( void )
//{
//}

/*
================
VID_IsFullScreen
================
*/
qboolean VID_IsFullScreen(void)
{
  return false;
}

static void
windowed_mouse_f(struct cvar_s *var)
{
    /*
    if (var->value) {
	Con_DPrintf("Callback: _windowed_mouse ON\n");
	if (!VID_IsFullScreen()) {
	    //IN_GrabMouse();
	    //IN_GrabKeyboard();
	}
    } else {
	Con_DPrintf("Callback: _windowed_mouse OFF\n");
	if (!VID_IsFullScreen()) {
	    //IN_UngrabMouse();
	    //IN_UngrabKeyboard();
	}
    }
    */
}

cvar_t _windowed_mouse = { "_windowed_mouse", "0", true, false, 0,
    windowed_mouse_f
};
